'use strict';

import { options, element, pageId, getPlayer } from './main.js';

const queue = [];
let queueRunning = false;
let requests = 0;
const searchTerms = {};
const timeout = 1000;

export function queueSegments(segments) {
	queue.push([...segments]);
	if (!queueRunning) {
		maybeRequestIdleCallback(requestSegments, { timeout });
	}
}

function requestSegments() {
	if (!queue.length) {
		document.body.classList.remove('vocabLoading');
		document.body.classList.add('vocabLoaded');
		queueRunning = false;
		return;
	}
	if (!queueRunning) {
		document.body.classList.add('vocabLoading');
		queueRunning = true;
	}
	const segments = queue.shift();
	const order = segments[0]?.dataset?.vocabBlock ?? '';
	const text = segments.map(node => '<div>' + node.innerHTML + '</div>').join('');
	const body = new URLSearchParams({
		text,
		chunk: pageId + '.' + order,
		language: options.lang || '',
		apiKey: options.apiKey || '',
	});
	startRequest();
	fetch(options.apiTermsUrl, {
		credentials: 'omit',
		method: 'POST',
		body,
	})
	.then(response => response.json())
	.then(data => {
		requestAnimationFrame(() => {
			for (let i = 0, j = segments.length; i < j; i++) {
				if (data.translations[i]) {
					segments[i].innerHTML = transformSegmentHtml(data.translations[i]);
				}
			}
		});
		maybeRequestIdleCallback(requestSegments, { timeout });
	})
	.finally(endRequest);
}

// return a Promise resolving to HTML
export function searchTermHTML(searchString = '', ptOfSpeech = '') {
	const searchKey = normalizeTerm(searchString) + '|' + ptOfSpeech;
	if (!(searchKey in searchTerms)) {
		startRequest();
		const body = new URLSearchParams({
			apikey: options.apiKey || '',
			searchLanguage: options.lang || '',
			searchString,
			ptOfSpeech,
		});
		searchTerms[searchKey] = fetch(options.apiSearchUrl, {
			credentials: 'omit',
			method: 'POST',
			body
		})
		.then(response => response.json())
		.then(data => {
			if (!data.success || !data.languages) return;
			const resultLanguage = Object.keys(data.languages)[0];
			if (!resultLanguage || !data.languages[resultLanguage]) return;
			let html = '';
			for (const result of data.languages[resultLanguage]) {
				html += result.results.map(r => r.src).join('');
			}
			return transformTermHtml(html);
		})
		.finally(endRequest);
	}
	return searchTerms[searchKey];

}

function normalizeTerm(term) {
	return term.toLowerCase();
}

function transformSegmentHtml(html) {
	return html.replaceAll('<mark', '<mark class="vocabLink"');
}

function transformTermHtml(html) {
	// dileu priodoleddau diangen
	// bwlch rhwng pob tag
	// cysoni enwau dosbarth (PT_class_name -> vocabClassName)
	return html.replace(/ (?:id|concept_id|xml:lang)=".*?"/g, '')
		.replace(/(<\/\w+>)\s*/g, '$1 ')
		.replace(/ class="(?:[A-Z]+_)?([^"]*)"/g, (m,n) =>
			n ? (' class="vocab' + n.replace(/(?:^|[_\-]+)(\w)/g, (o,p) => p.toUpperCase()) + '"') : ''
		);
}

export function synthesize(testun, siaradwr) {
	siaradwr ??= options.speaker;
	const body = new URLSearchParams({
		testun, siaradwr,
		api_key: options.apiKey || '',
	});
	startRequest();
    fetch(options.apiSpeechUrl, {
		credentials: 'omit',
		method: 'POST',
		body,
	}).then(res => {
        if (!res.ok) {
			throw Error(res.statusText);
		}
        return res.blob();
    }).then(blob => {
		const player = getPlayer();
        player.src = URL.createObjectURL(blob);
		player.play();
    }).finally(endRequest);
}

function startRequest() {
	if (!requests++) {
		element.dispatchEvent(new CustomEvent('vocabState', {
			detail: { busy: true }
		}));
	}
}

function endRequest() {
	if (requests && !--requests) {
		element.dispatchEvent(new CustomEvent('vocabState', {
			detail: { busy: false }
		}));
	}
}

// Polyfill for browsers without support
function maybeRequestIdleCallback(callback, options) {
	if ('requestIdleCallback' in window) {
		requestIdleCallback(callback, options);
	} else {
		setTimeout(callback, 0);
	}
}
