'use strict';

import { insertMenu } from './bar.js';
import { showPopup, hidePopup, hasPopup } from './popup.js';
import { getTextNodesIn } from './text.js';
import { queueSegments, searchTermHTML, synthesize } from './api.js';

export const options = {
	lang: 'cy',
	delay: 100,
	ignore: '.vocabIgnore',
	stylesheet: new URL('main.css', document.currentScript?.src || window.location),
	apiTermsUrl: 'https://api.techiaith.org/porthtermau/api/v1/vocab',
	apiSearchUrl: 'https://api.techiaith.org/porthtermau/api/v1/vocabsearch',
	apiSpeechUrl: 'https://api.techiaith.org/coqui-tts/api/v1',
	build: typeof WEBPACK_BUILD === 'undefined' ? Date.now() : WEBPACK_BUILD,
};

export const isTouch = !!('ontouchstart' in window) || window.navigator.msMaxTouchPoints > 0;
export let link = null;
export let element = null;
export let pageId = null;

let loaded = false;
let active = false;
let timer = null;
let observer;
const blocks = [];
const requestedBlocks = [];
const maxBlockSize = 2048;

export function init(userOptions = {}) {
	Object.assign(options, userOptions);
	if (!loaded) {
		element = (options.start && document.querySelector(options.start)) || document.body;
		observer = new IntersectionObserver(onIntersection, {
			rootMargin: "100px",
		});
		pageId = Date.now() + '.' + hash(window.location.href);
		attachSegments(element);
		insertStyles(options.stylesheet);
		insertMenu();
		document.body.addEventListener('keydown', onKey);
		document.body.classList.add('vocabTheme', isTouch ? 'vocabTouch' : 'vocabPointer');
		element.addEventListener('vocabState', onStateChange);
		document.addEventListener('selectionchange', onSelection);
		loaded = true;
		debug('vocab wedi llwytho', element);
	}
}

export function getOption(name) {
	return options[name];
}

export function isActive() {
	return active;
}

export function turnOn() {
	if (!loaded) init();
	element.addEventListener('mousemove', onMouseMove);
	element.addEventListener('vocabContentLoaded', onContent);
	document.body.classList.add('vocabActive');
	options.button.setAttribute('aria-checked', 'true');
	active = true;
	debug('vocab ymlaen');
}

export function turnOff() {
	element.removeEventListener('mousemove', onMouseMove);
	element.removeEventListener('vocabContentLoaded', onContent);
	document.body.classList.remove('vocabActive');
	options.button.setAttribute('aria-checked', 'false');
	hidePopup();
	active = false;
	debug('vocab i ffwrdd');
}

export function toggle() {
	active ? turnOff() : turnOn();
	return active;
}

export function showPrevious() {
	showPreviousNext(true);
}

export function showNext() {
	showPreviousNext(false);
}

function showPreviousNext(isPrevious) {
	const links = Array.from(element.querySelectorAll('.vocabLink'));
	if (isPrevious) links.reverse();
	const pos = (link ? links.indexOf(link) : -1) + 1;
	for (const link of links.slice(pos)) {
		const box = link.getBoundingClientRect();
		if (!box.width || !box.height) continue;
		link.scrollIntoView({block:"center"});
		activateLink(link);
		break;
	}
}

function onStateChange(e) {
	document.body.classList.toggle('vocabLoading', e.detail.busy);
}

function onKey(e) {
	if (e.shiftKey || e.metaKey) return;
	if (e.keyCode === 86 && ((e.altKey && e.ctrlKey) || (active && !e.altKey && !e.ctrlKey))) {
		e.preventDefault();
		active ? turnOff() : turnOn();
		return;
	}
	const keyActions = {
		' ': togglePopup,
		'Escape': turnOff,
		'ArrowLeft': showPrevious,
		'ArrowRight': showNext,
	}
	if (active && e.key in keyActions) {
		e.preventDefault();
		keyActions[e.key]();
	}
}

function onMouseMove(e) {
	if (link !== e.target) {
		if (e.target.classList.contains('vocabLink')) {
			activateLink(e.target);
		} else if (link) {
			deactivateLink();
		}
	}
}

function onContent(e) {
	if (e.detail.link === link) {
		showPopup(e.detail.link, e.detail.content);
		if (options.audio) {
			synthesize(link.textContent);
		}
	}
}

function onSelection(e) {
	const text = document.getSelection()?.toString();
	debug(e, document.getSelection(), text);
	if (text && options.audio) {
		debounce(() => synthesize(text), 500);
	}
}

function activateLink(newLink) {
	if (newLink !== link) {
		if (link) deactivateLink();
		link = newLink;
		link.classList.add('vocabActiveLink');
		timer = setTimeout(loadPopup, options.delay);
	}
}

export function deactivateLink() {
	clearTimeout(timer);
	link?.classList.remove('vocabActiveLink');
	link = null;
	hidePopup();
}

function loadPopup() {
	if (link) {
		const thisLink = link;
		searchTermHTML(link.textContent, link.dataset.pos).then(content => {
			if (content) {
				element.dispatchEvent(new CustomEvent('vocabContentLoaded', {
					detail: { link: thisLink, content }
				}));
			} else {
				debug('Dim canlyniad', link);
				unwrap(link);
			}
		});
	}
}

function togglePopup() {
	if (hasPopup()) {
		hidePopup();
	} else {
		loadPopup();
	}
}

function attachSegments(context) {
	let blockNumber = 0;
	let blockSize = 0;
	for (const node of getTextNodesIn(context)) {
		if (!/\S/.test(node.textContent.match)) continue; // ignore just whitespace
		const span = document.createElement('span');
		node.after(span);
		span.appendChild(node);
		const size = span.innerHTML.length;
		if (blocks[blockNumber]) {
			if (blockSize + size > maxBlockSize) {
				blocks[++blockNumber] = [];
				blockSize = 0;
			}
			blocks[blockNumber].push(span);
		} else {
			blocks[blockNumber] = [span];
		}
		blockSize += size;
		span.dataset.vocabBlock = blockNumber;
		observer.observe(span);
	}
	return blocks;
}

function onIntersection(entries, observer) {
	entries.forEach((entry) => {
		if (entry.isIntersecting) {
			const blockNumber = entry.target.dataset.vocabBlock;
			if (blockNumber && !requestedBlocks[+blockNumber]) {
				requestedBlocks[+blockNumber] = true;
				queueSegments(blocks[blockNumber]);
			}
			observer.unobserve(entry.target);
		}
	});
}

function debounce(func, duration, immediate) {
	let timeout;
	return function (...args) {
		if (immediate && !timeout) {
			func.apply(this, args);
		}
		const effect = () => {
			timeout = null;
			if (!immediate) func.apply(this, args);
		};
		clearTimeout(timeout);
		timeout = setTimeout(effect, duration);
	}
}

export function getPlayer() {
	return options.player;
}

function insertStyles(href) {
	if (href) {
		document.head.insertAdjacentHTML('beforeend',
			`<link rel="stylesheet" media="screen" href="${href}?v=${options.build}"/>`
		);
	}
}

function hash(str) {
	let hash = 0
	for (let i = 0; i < str.length; i++) {
		hash = ((hash << 5) - hash + str.charCodeAt(i)) | 0
	}
	return (hash >>> 0).toString(36)
}

function unwrap(el) {
	requestAnimationFrame(() => {
		while (el.firstChild) {
			el.parentNode.insertBefore(el.firstChild, el);
		}
	});
}

function debug(...args) {
	if (options.debug || window.location.search.includes('vocabDebug')) {
		console.debug(...args);
	}
}